import java.awt.Color;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JApplet;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.Popup;
import javax.swing.PopupFactory;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.JLabel;

public class Controller extends JApplet {

	private static final long serialVersionUID = 1L;

	final int board_width = 5;
	final int player_count = 6;
	//the main console label (also as a sentinel to the "last component bug")
	JLabel label = null;
	
	//Game board and players
	//ArrayList<Field[][]> history = new ArrayList<Controller.Field[][]>();
	Field[][] board = new Field[board_width*2][board_width*2];
	Player[] players = new Player[player_count];
	JLabel[] player_label = new JLabel[player_count];
	JPanel player_panel = null;
	
	ImageIcon icon_dead;
	
	//Who is on turn and what field is clicked
	Point actual_point = null;
	int actual_player_id = 0;
	int round_number = 1;
	
	int start_x = 50;
	int start_y = 0;
	
	//[0] is looser, ...
	ArrayList<Player> player_order = new ArrayList<Controller.Player>(); 
	
	//Application start up
	public void init()
	{
		try{
			SwingUtilities.invokeAndWait(new Runnable(){
				public void run() {
					createGUI();
				}
			});
		} catch (Exception e){
			e.printStackTrace();
		}
	}
	
	//http://download.oracle.com/javase/tutorial/uiswing/components/icon.html
	protected ImageIcon createImageIcon(String path, int width, int height) {
		ImageIcon result = null;
		
		try{
		    ImageIcon imgico = new ImageIcon(this.getClass().getResource(path));
		    result = new ImageIcon(imgico.getImage().getScaledInstance(width, height, java.awt.Image.SCALE_SMOOTH));
		}catch(Exception e){
			e.printStackTrace();
		}
		
		return result; 
	}

	//If this coordinates are on the gameboard (it is a hex game)
	private boolean isInBoard(int i, int j)
	{
		return !(
				((i+j)<board_width-1) 
				|| ((i+j)>board_width*3 - 3)
				|| i<0
				|| j<0
				|| i>board_width*2-2
				|| j>board_width*2-2
		);
	}

	//Booleans which specify that the game has ended
	//foreach stone of Player p check if it can move
	private boolean canMovePlayer(Player p)
	{
		for(int i=0; i<board.length ; i++){
			for(int j=0; j<board[i].length ; j++){
				if(isInBoard(i, j))
					if(board[i][j].getOwner() == p){
						if(board[i][j].canMove()){
							return true;
						}
				}
			}
		}
		return false;
	}
	
	//!ASSUMPTION --- if it is not full, then someone could move!
	//If in all board components some player starts then the assumption is true
	private boolean isBoardFull(){
		for(int i=0; i<board.length ; i++){
			for(int j=0; j<board[i].length ; j++){
				if(isInBoard(i,j)){
					if(board[i][j].getFieldType() == FieldType.NOTHING){
						return false;
					}
				}
			}
		}
		return true;
	}
	
	//returns how many players are still alive
	int playersAliveCount()
	{
		int result = 0;
		for(int i = 0; i < players.length ; i++){
			if(!players[i].is_dead){
				result++;
			}
		}
		return result;
	}
	
	//Count how many stones has which player
	void actualisePlayersStoneCount()
	{
		for(int i=0; i<players.length ;i++){
			players[i].stone_count=0;
		}
		for(int i=0; i<board.length; i++)
			for(int j=0; j<board[i].length; j++)
				if(isInBoard(i,j))
					if(board[i][j].getFieldType() == FieldType.STONE)
						board[i][j].getOwner().stone_count++;
		
		for(int i=0; i<players.length ;i++){
			player_label[i].setText(((Integer)players[i].stone_count).toString());
		}
	}
	
	//Erases this player from game (his stones turn to "obstacles")
	void setPlayerDead(Player p)
	{
		if(p.is_dead) return;
		
		for(int i=0; i<board.length; i++)
			for(int j=0; j<board[i].length; j++)
				if(isInBoard(i,j))
					if(board[i][j].getFieldType() == FieldType.STONE && board[i][j].getOwner() == p){
						board[i][j].setDead();
					}
		p.is_dead = true;
		player_order.add(p);
		actualisePlayersStoneCount();
	}
	
	//Adds rest of players to order (results) of the game
	//Yes, it is a very lame implementation
	void restPlayersToOrder()
	{
		for(int i=0; i<players.length ;i++){
			int min = 123456789;
			int min_player_id = -1;
			for(int j=0; j<players.length; j++){
				if(0 < players[j].stone_count && players[j].stone_count < min && !players[j].is_dead){
					min = players[j].stone_count;
					min_player_id = j;
				}
			}
			
			if(min_player_id >= 0){
				players[min_player_id].is_dead = true;
				player_order.add(players[min_player_id]);
			}
		}
	}
	
	void killZeroStonePlayers()
	{
		actualisePlayersStoneCount();
		for(int i=0; i<players.length ;i++){
			if(players[i].stone_count == 0){
				setPlayerDead(players[i]);
			}
		}
	}
	
	void showWinnerPopup()
	{
		JPanel panel = new JPanel();
		panel.setBounds(0, 0, 150, 350);
		panel.setBorder(BorderFactory.createLineBorder(new Color(0,0,0), 3));
		panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
		
	    final Popup popup = PopupFactory.getSharedInstance().getPopup(this, panel, 100, 100);
	    
	    JLabel main_text = new JLabel("Game ended after " + ((Integer)round_number).toString()+ "rounds.", SwingConstants.CENTER);
	    main_text.setBorder(BorderFactory.createMatteBorder(2, 2, 2, 2, new Color(255,0,0)));
		panel.add(main_text);
		for(int i=0; i<player_order.size() ; i++){
			JLabel l = new JLabel(
					((Integer)(players.length - i)).toString() + 
					". with " 
					+ ((Integer)player_order.get(i).stone_count).toString() + " stones", 
							player_order.get(i).icon, SwingConstants.CENTER
			);
			panel.add(l);
			l.setBorder(BorderFactory.createLineBorder(new Color(0,0,0), 1));
			l.setHorizontalAlignment(SwingConstants.CENTER);
		} 
		
		ActionListener hider = new ActionListener() {
	        public void actionPerformed(ActionEvent e) {
	            popup.hide();
	          }
	    };
		JButton button = new JButton();
		button.addActionListener(hider);
		button.setText("OK");
		
		//this.removeAll();
		this.add(panel);
		/*
		panel.add(button);
		try{
		  popup.show();
		}catch(Exception e){
			e.printStackTrace();
		}*/
	}
	
	//after player make his move, next player is on turn (some logic must be updated) 
	//also the winner / looser mechanism is here implemented 
	void nextPlayer()
	{
		//nothing to do in this game
		if(playersAliveCount() == 0){
			return;
		}
		
		actualisePlayersStoneCount();
		int winner_id = -1;
		
		//unset that I`m the actual
		player_label[actual_player_id].setBorder(BorderFactory.createLineBorder(new Color(255,255,255)));
		
		//who as at most stones, he wins
		if(isBoardFull()){
			int max_stone = -1;
			for(int i=0; i<players.length ;i++){
				//if same stone count, decides who started later
				if(players[i].stone_count >= max_stone){
					max_stone = players[i].stone_count;
					winner_id = i;
				}
			}
		}
		if(playersAliveCount() == 1){
			for(int i=0; i < players.length ; i++){
				if(!players[i].is_dead)
					winner_id = i;
			}
		}
		
		if(winner_id >= 0){
			killZeroStonePlayers();
			restPlayersToOrder();
			showWinnerPopup();
			return;
		}

		//safety before infinite loop
		if(playersAliveCount() == 0){
			return;
		}
			
		//while !is_dead that player
		do{
		  actual_player_id = (actual_player_id + 1) % players.length;
		  if(actual_player_id == 0) round_number++;
		}while(players[actual_player_id].is_dead);
		
		Player actual_player = players[actual_player_id];
		player_label[actual_player_id].setBorder(BorderFactory.createLineBorder(new Color(255,0,0)));
		
		//loose only if someone is still alive
		if(!canMovePlayer(actual_player) || actual_player.stone_count==0){
			setPlayerDead(actual_player);
			nextPlayer();
		}
	}
	
	private void createGUI()
	{	
		for(int i=0; i<board.length; i++)
			for(int j=0; j<board[i].length; j++)
				board[i][j] = null;
		
		int field_alloc_size = 60;
		int field_true_size = 50;
		double board_size_modifier = 2.0;
		setSize(
				(int)((double)(field_alloc_size*board_width)*board_size_modifier) + 150, 
				(int)((double)(field_alloc_size*board_width)*board_size_modifier)
			);
		
		try{
			players[0] = new Player("Petrz", createImageIcon("black.png", field_true_size, field_true_size));
			players[1] = new Player("Bubi", createImageIcon("blue.png", field_true_size, field_true_size));
			players[2] = new Player("Kombo", createImageIcon("green.png", field_true_size, field_true_size));
			players[3] = new Player("Rizo", createImageIcon("purple.png", field_true_size, field_true_size));
			players[4] = new Player("Beren", createImageIcon("red.png", field_true_size, field_true_size));
			players[5] = new Player("Edo", createImageIcon("yellow.png", field_true_size, field_true_size));
			
			icon_dead = createImageIcon("dead.png", field_true_size, field_true_size);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		player_panel = new JPanel();
		int pp_width = 500;
		int pp_height = 100;
		player_panel.setBounds(getSize().width-pp_width, getSize().height - pp_height, pp_width, pp_height);
		add(player_panel);
		
		for(int i=0; i<players.length ;i++){
			player_label[i] = new JLabel(players[i].name, players[i].icon, SwingConstants.RIGHT);
			player_label[i].setText(((Integer)players[i].stone_count).toString());
			player_label[i].setBorder(BorderFactory.createLineBorder(new Color(255,255,255)));
			player_label[i].setBounds(0, i*50,pp_width, 50);
			player_panel.add(player_label[i]);
		}
		
		//which fields are in my neighbourhood (expressed in points)
		Point[] neighbourhood = getNeighbourhood();
		
		//creates board
		for(int i=board_width*2-2; i>=0; i--){
			for(int j=board_width*2-2; j>=0; j--){
				if(!isInBoard(i,j)) continue;
				
				Field help = new Field(null, FieldType.NOTHING, new Point(i,j));
				add(help);
				board[i][j] = help;
				/*
				if(i==0 || j==0 || ((i+j)<=board_width-1) || ((i+j)>=board_width*3 - 3))
					help = new Field(null, FieldType.BORDER);
				else 
					help = new Field(null, FieldType.NOTHING);*/
				
				help.setBounds(
						start_x + (int)((double)field_alloc_size*((double)i+((double)j)*0.5)), 
						start_y + (int)((double)field_alloc_size*(((double)j)*Math.sqrt(3.0)*0.5)), 
						field_true_size, 
						field_true_size
				);
				
				help.addMouseListener(new FieldMouseEventHandler());
				//help.addKeyListener(new FieldKeyEventHandler());
				//help.setBackground(new Color(255,128,128));
				//help.setOpaque(true);
				//help.setText(((Integer)i).toString() + "," + ((Integer)j).toString());
				help.setBorder(BorderFactory.createLineBorder(new Color(0,0,0)));
				
				
				for(int n=0; n<neighbourhood.length ; n++)
				{
					int a = neighbourhood[n].x + i;
					int b = neighbourhood[n].y + j;
					
					if(!isInBoard(a, b)) continue;
					if(board[a][b] == null) continue;
					
					board[i][j].addNeighbour(board[a][b]);
					board[a][b].addNeighbour(board[i][j]);
				}
			}
		}

		//Starting position
		board[board_width-1][0].setOwner(players[0]);
		board[2*(board_width-1)][0].setOwner(players[1]);
		board[2*(board_width-1)][board_width-1].setOwner(players[2]);
		board[board_width-1][2*(board_width-1)].setOwner(players[3]);
		board[0][2*(board_width-1)].setOwner(players[4]);
		board[0][board_width-1].setOwner(players[5]);

		//!TODO -- last added components bounds are never considered in parent component
		label = new JLabel();
		//label.setBounds(0,0,5,5);
		add(label);
		//label.setText(board[0][9].getLocation().toString() + " " + board[0][9].getSize().toString());
		actualisePlayersStoneCount();
	}

	//Returns where I can move by one
	public Point[] getNeighbourhood()
	{
		Point[] neighbourhood = new Point[6]; 
		neighbourhood[0] = new Point(1,0); 
		neighbourhood[1] = new Point(-1,0); 
		neighbourhood[2] = new Point(0,1); 
		neighbourhood[3] = new Point(0,-1); 
		neighbourhood[4] = new Point(-1,1); 
		neighbourhood[5] = new Point(1,-1);
		return neighbourhood;
	}
		
	//Represents a field on board, all data is kept here (also the JLabel data)
	public class Field extends JLabel
	{
		private static final long serialVersionUID = 2L;
		
		public Field(Player _owner, FieldType _type, Point _board_point)
		{
			this.owner = _owner;
			this.type = _type;
			this.board_point = _board_point;
			this.neighbours = new ArrayList<Controller.Field>();
		}
		public Player getOwner() {return owner;}
		public FieldType getFieldType() {return type;}
		public Point getBoardPoint() {return board_point;}
		
		public void setOwner(Player p)
		{
			owner = p;
			type = FieldType.STONE;
			this.setIcon(p.icon);
		}
		public void setNothing()
		{
			this.owner = null;
			this.type = FieldType.NOTHING;
			this.setIcon(null);
		}
		public void setDead()
		{
			this.owner = null;
			this.type = FieldType.DEAD;
			this.setIcon(icon_dead);
		}
		
		private Player owner;
		private FieldType type;
		
		public void addNeighbour(Field f) {this.neighbours.add(f);}
		public void setNeighbourAsI(){
			if(type != FieldType.STONE) return;
			for(int i=0; i<neighbours.size() ;i++){
				if(neighbours.get(i).getOwner() == null) continue;
				neighbours.get(i).setOwner(owner);
			}
		}
	
		public boolean isNeighbour(Field f){
			for(int i=0; i<this.neighbours.size() ; i++){
				if(this.neighbours.get(i).equals(f)) return true;
			}
			return false;
		}
		//DFS
		public boolean isNeighbourInDist(Field f, int dist){
			if(dist == 0) return (f.equals(this));
			if(dist == 1) return isNeighbour(f);
			
			for(int i=0; i<this.neighbours.size() ; i++){
				if(this.neighbours.get(i).isNeighbourInDist(f, dist-1)) return true;
			}
			return false;
		}
		public boolean isFreeNeighbour(/*dist = 1*/){
			for(int i=0; i<this.neighbours.size() ; i++){
				if(this.neighbours.get(i).getFieldType() == FieldType.NOTHING) return true;
			}
			return false;
		}
		public boolean isFreeNeighbourInDist(int dist){
			if(dist == 1) return isFreeNeighbour();
			
			for(int i=0; i<this.neighbours.size() ; i++){
				if(this.neighbours.get(i).isFreeNeighbourInDist(dist-1)) return true;
			}
			return false;
		}
		
		//returns true if: (not stone) or (stone and cold move somewhere)
		public boolean canMove()
		{
			return this.isFreeNeighbourInDist(2);
		}
		
		private ArrayList<Field> neighbours;
		private Point board_point;
		
	}
	
	//Types of fields
	//Assumption: just NOTHING is a free field
	public static enum FieldType
	{
		STONE,
		HOLE,
		NOTHING,
		DEAD
	}
	
	//just a Player struct
	public class Player
	{
		public Player(String _name, Icon _icon)
		{
			name = _name;
			icon = _icon;
		}
		public String name;
		public Icon icon;
		public int stone_count = 0;
		public boolean is_dead = false;
	}
    
	//onClicked event - actual_player_id and move to
	public class FieldMouseEventHandler implements MouseListener {

		@Override
		public void mouseClicked(MouseEvent arg0) {
			//Just a safety rule
			if(!(arg0.getSource() instanceof Field)) return;

			Field f = (Field) arg0.getSource();
			Player p = players[actual_player_id];
			if(actual_point == null){
				//you can select only your own stone
				if(f.getFieldType() == FieldType.STONE && f.getOwner() == p){
					actual_point = f.getBoardPoint();
					f.setBorder(BorderFactory.createLineBorder(new Color(255,0,0)));
				}
			}
			else{
			  Field f_act = board[actual_point.x][actual_point.y];
			  
			  //clicked on himself
			  if(f == f_act){
				  actual_point = null;
			  }
			  //clicked on free neighbour in radius 1
			  else if(f.isNeighbour(f_act) && f.getFieldType() == FieldType.NOTHING){
				  f.setOwner(f_act.getOwner());
				  f.setNeighbourAsI();
				  nextPlayer();
			//clicked on nieghbour in radius 2
			  }else if(f.isNeighbourInDist(f_act, 2) && f.getFieldType() == FieldType.NOTHING){
				  f.setOwner(f_act.getOwner());
				  f_act.setNothing();
				  f.setNeighbourAsI();
				  nextPlayer();
			//something else as a new click without actual_point
			  }else{
				  actual_point = null;
				  mouseClicked(arg0);
			  }
			  //Odvyznacit
			  actual_point = null;
			  f_act.setBorder(BorderFactory.createLineBorder(new Color(0,0,0)));
			  
			}
		}

		@Override
		public void mouseEntered(MouseEvent arg0) {
			//if(!(arg0.getSource() instanceof Field)) return;
			//label.setText("Entered " + arg0.getSource());
		}

		@Override
		public void mouseExited(MouseEvent arg0) {
			//if(!(arg0.getSource() instanceof Field)) return;
			//label.setText("Exited " + arg0.getSource());
		}

		@Override
		public void mousePressed(MouseEvent arg0) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void mouseReleased(MouseEvent arg0) {
			// TODO Auto-generated method stub
			
		}
		
	}
	
	/*
	public class FieldKeyEventHandler implements KeyListener
	{

		@Override
		public void keyPressed(KeyEvent arg0) {
			// TODO Auto-generated method stub
			if(arg0.getKeyCode() == KeyEvent.VK_BACK_SPACE){
				if(history.size() > 0){
					board = history.get(history.size() - 1);
					history.remove(history.size() - 1);
				}
			}
		}

		@Override
		public void keyReleased(KeyEvent arg0) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void keyTyped(KeyEvent arg0) {
			// TODO Auto-generated method stub
			
		}
		
	}*/

}